/*
 * Copyright 2020 University of Liège
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "wPicard.h"
#include "wProblem.h"
#include "wMedium.h"
#include "wAssign.h"
#include "wFreestream.h"
#include "wWake.h"
#include "wKutta.h"
#include "wBlowing.h"
#include "wPotentialResidual.h"
#include "wFreestreamResidual.h"
#include "wWakeResidual.h"
#include "wKuttaResidual.h"
#include "wBlowingResidual.h"

#include "wMshData.h"
#include "wTag.h"
#include "wNode.h"
#include "wElement.h"
#include "wLinearSolver.h"

#include <Eigen/Sparse>

#include <tbb/global_control.h>
#include <tbb/parallel_for_each.h>
#include <tbb/spin_mutex.h>

#include <iomanip>
#include <deque>

using namespace tbox;
using namespace flow;

#define ANSI_COLOR_RED "\x1b[1;31m"
#define ANSI_COLOR_GREEN "\x1b[1;32m"
#define ANSI_COLOR_YELLOW "\x1b[1;33m"
#define ANSI_COLOR_RESET "\x1b[0m"

/**
 * @brief Initialize Picard solver
 * @authors Adrien Crovato
 */
Picard::Picard(std::shared_ptr<Problem> _pbl, std::shared_ptr<tbox::LinearSolver> _linsol) : Solver(_pbl, _linsol)
{
    // Additional default parameters
    relax = 1.;
}

/**
 * @brief Run the Picard sovler
 *
 * Solve the steady subsonic Full Potential Equation
 * @authors Adrien Crovato
 */
bool Picard::run()
{
    // Multithread
    tbb::global_control control(tbb::global_control::max_allowed_parallelism, nthreads);

    // Display solver and problem parameters
    std::cout << "--- Picard solver ---\n"
              << "Inner solver: " << *linsol
              << std::setprecision(5)
              << "Number of threads: " << nthreads << "\n"
              << "Relative tolerance: " << log10(relTol) << "\n"
              << "Absolute tolerance: " << log10(absTol) << "\n"
              << "Maximum iterations: " << maxIt << "\n";
    std::cout << "--- Problem definition ---\n"
              << std::setprecision(5)
              << "Angle of attack: " << pbl->alpha * 180 / 3.14159 << " deg\n"
              << "Angle of sideslip: " << pbl->beta * 180 / 3.14159 << " deg\n"
              << "Mach number: " << pbl->M_inf << "\n"
              << "Reference area: " << pbl->S_ref << "\n"
              << "Reference length: " << pbl->c_ref << "\n"
              << "Reference origin: (" << pbl->x_ref(0) << ", " << pbl->x_ref(1) << ", " << pbl->x_ref(2) << ")\n"
              << std::endl;

    // Initialize solver loop
    nIt = 0;
    double resInit = 1.;
    double absRes = 1., relRes = 1.;
    // Solution, residual and old solution
    Eigen::Map<Eigen::VectorXd> phi_(phi.data(), phi.size()), rPhi_(rPhi.data(), rPhi.size());
    Eigen::VectorXd phiOld(phi.size());

    std::cout << std::setw(8) << "N_Iter"
              << std::setw(12) << "Cl"
              << std::setw(12) << "Cd"
              << std::setw(15) << "Rel_Res[phi]"
              << std::setw(15) << "Abs_Res[phi]" << std::endl;

    do
    {
        // Reinitialize SoE
        Eigen::SparseMatrix<double, Eigen::RowMajor> A(pbl->msh->nodes.size(), pbl->msh->nodes.size());
        std::vector<double> b(pbl->msh->nodes.size()); // dummy b vector
        Eigen::Map<Eigen::VectorXd> b_ = Eigen::Map<Eigen::VectorXd>(b.data(), b.size());

        // Build system of equations A \phi = b
        build(A, b);

        // Compute residual
        rPhi_ = A * phi_ - b_;
        if (nIt == 0)
            resInit = rPhi_.norm();
        absRes = rPhi_.norm();
        relRes = absRes / resInit;

        // Compute aerodynamic coefficient on boundaries
        Solver::computeLoad();

        // Display residual
        std::cout << std::fixed << std::setprecision(5);
        std::cout << std::setw(8) << nIt
                  << std::setw(12) << Cl
                  << std::setw(12) << Cd
                  << std::setw(15) << log10(relRes)
                  << std::setw(15) << log10(absRes) << "\n";

        // Check convergence
        if ((relRes < relTol || absRes < absTol) && nIt != 0)
            break;
        else if (std::isnan(relRes))
            break;
        nIt++;

        // Transfer old solution
        phiOld = phi_;

        // Solve linear SoE
        tms["4-Slinsol"].start();
        linsol->compute(A, b_, phi_);
        tms["4-Slinsol"].stop();

        // Relaxation
        phi_ = (1 - relax) * phiOld + relax * phi_;
    } while (nIt < maxIt);

    // Compute field variables
    Solver::computeFlow();

    // Check the solution
    if (relRes < relTol || absRes < absTol)
    {
        std::cout << ANSI_COLOR_GREEN << "Picard solver converged!" << ANSI_COLOR_RESET << std::endl;
        if (verbose > 0)
            std::cout << "Picard solver CPU" << std::endl
                      << tms;
        std::cout << std::endl;
        return true;
    }
    else if (std::isnan(relRes))
    {
        std::cout << ANSI_COLOR_RED << "Picard sovler diverged!" << ANSI_COLOR_RESET << std::endl;
        if (verbose > 0)
            std::cout << "Picard solver CPU" << std::endl
                      << tms;
        std::cout << std::endl;
        return false;
    }
    else
    {
        std::cout << ANSI_COLOR_YELLOW << "Picard solver not fully converged!" << ANSI_COLOR_RESET << std::endl;
        if (verbose > 0)
            std::cout << "Picard solver CPU" << std::endl
                      << tms;
        std::cout << std::endl;
        return true;
    }
}

/**
 * @brief Build LHS matrix and RHS vector for Picard iteration
 * @authors Adrien Crovato
 */
void Picard::build(Eigen::SparseMatrix<double, Eigen::RowMajor> &A, std::vector<double> &b)
{
    // Multithread
    tbb::spin_mutex mutex;

    // List of triplets to build matrix
    std::deque<Eigen::Triplet<double>> T;

    // Full Potential Equation: analytical derivatives
    tms["0-Abase"].start();
    auto fluid = pbl->medium;
    tbb::parallel_for_each(fluid->adjMap.begin(), fluid->adjMap.end(), [&](std::pair<Element *, std::vector<Element *>> p) {
        // Current element
        Element *e = p.first;

        // Elementary stiffness
        Eigen::MatrixXd Ae = PotentialResidual::buildFixed(*e, phi, *fluid);

        // Assembly
        tbb::spin_mutex::scoped_lock lock(mutex);
        for (size_t ii = 0; ii < e->nodes.size(); ++ii)
        {
            Node *nodi = e->nodes[ii];
            for (size_t jj = 0; jj < e->nodes.size(); ++jj)
            {
                Node *nodj = e->nodes[jj];
                T.push_back(Eigen::Triplet<double>(rows[nodi->row], nodj->row, Ae(ii, jj)));
            }
        }
    });
    tms["0-Abase"].stop();

    // Kutta condition step 2: equality of velocity norm through the wake
    // -> add the Kutta equation on the upper wake node lines
    // Nishida Thesis, pp. 29-30, Eq. 2.18
    tms["1-Akutta"].start();
    for (auto wake : pbl->wBCs)
    {
        tbb::parallel_for_each(wake->wEle.begin(), wake->wEle.end(), [&](WakeElement *we) {
            // Build Ku, Kl
            Eigen::MatrixXd Aue, Ale;
            std::tie(Aue, Ale) = WakeResidual::buildFixed(*we, phi);
            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < we->nRow; ++ii)
            {
                Node *nodi = we->wkN[ii].second;
                for (size_t jj = 0; jj < we->nColUp; ++jj)
                {
                    Node *nodj = we->volUpE->nodes[jj];
                    T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, Aue(ii, jj)));
                }
                for (size_t jj = 0; jj < we->nColLw; ++jj)
                {
                    Node *nodj = we->volLwE->nodes[jj];
                    T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, -Ale(ii, jj)));
                }
            }
        });
    }
    // Kutta condition step 3: equality of velocity norm on the Trailing edge (Kutta)
    // -> add the Kutta equation on the upper TE node lines (only works for 2D cases)
    // Galbraith paper, pp. 7, Eq. 34
    for (auto kutta : pbl->kSCs)
    {
        tbb::parallel_for_each(kutta->kEle.begin(), kutta->kEle.end(), [&](KuttaElement *ke) {
            // Build K
            Eigen::MatrixXd Ae = KuttaResidual::buildFixed(*ke, phi);

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < ke->nRow; ++ii)
            {
                Node *nodi = ke->teN[ii].second;
                for (size_t jj = 0; jj < ke->nCol; ++jj)
                {
                    Node *nodj = ke->volE->nodes[jj];
                    T.push_back(Eigen::Triplet<double>(nodi->row, nodj->row, Ae(ii, jj)));
                }
            }
        });
    }
    tms["1-Akutta"].stop();

    // Fill vector b: loop on all Neumann boundaries (except slip condition which are naturally enforced)
    tms["2-bnbc"].start();
    for (auto nBC : pbl->fBCs)
    {
        tbb::parallel_for_each(nBC->tag->elems.begin(), nBC->tag->elems.end(), [&](Element *e) {
            // Build elementary flux vector
            Eigen::VectorXd be = FreestreamResidual::build(*e, phi, *nBC);

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < e->nodes.size(); ++ii)
            {
                Node *nodi = e->nodes[ii];
                b[rows[nodi->row]] -= be(ii);
            }
        });
    }
    // Apply blowing (transpiration) BCs
    for (auto bBC : pbl->bBCs)
    {
        tbb::parallel_for_each(bBC->uE.begin(), bBC->uE.end(), [&](std::pair<Element *, F0ElC *> p) {
            // Build elementary flux vector
            Eigen::VectorXd be = BlowingResidual::build(*p.first, phi, *p.second);

            // Assembly
            tbb::spin_mutex::scoped_lock lock(mutex);
            for (size_t ii = 0; ii < p.first->nodes.size(); ++ii)
            {
                Node *nodi = p.first->nodes[ii];
                b[rows[nodi->row]] -= be(ii);
            }
        });
    }
    tms["2-bnbc"].stop();

    // Build matrix without BCs
    A.setFromTriplets(T.begin(), T.end());

    // Apply Dirichlet BCs to A matrix and b vector
    tms["3-Abdbc"].start();
    for (auto dBC : pbl->dBCs)
    {
        for (auto nod : dBC->nodes)
        {
            for (Eigen::SparseMatrix<double, Eigen::RowMajor>::InnerIterator it(A, nod->row); it; ++it)
            {
                if (it.row() == it.col())
                    it.valueRef() = 1.;
                else
                    it.valueRef() = 0.;
            }
        }
        dBC->apply(b);
    }
    tms["3-Abdbc"].stop();

    // Clean matrix and turn to compressed row format
    A.prune(0.);
    A.makeCompressed();

    if (verbose > 1)
    {
        std::cout << "A (" << A.rows() << "," << A.cols() << ") nnz=" << A.nonZeros() << "\n";
        std::cout << "b (" << b.size() << ")\n";
    }
}

void Picard::write(std::ostream &out) const
{
    out << "flow::Picard"
        << "\n";
}
