#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright 2020 University of Liège
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


## Compute the flow around a NACA 0012 wing on a deformed mesh
# Adrien Crovato
#
# Test the mesh deformation process
#
# CAUTION
# This test is provided to ensure that the solver works properly.
# Mesh refinement may have to be performed to obtain physical results.

import numpy as np
import flow.utils as floU
import flow.default as floD
import fwk
from fwk.testing import *
from fwk.coloring import ccolors

def main():
    # timer
    tms = fwk.Timers()
    tms['total'].start()

    # define flow variables
    alpha0 = 0 # must be zero for this testfile
    M_inf = 0.3
    dim = 3

    # define dimension and mesh size
    spn = 1.0 # wing span
    lgt = 3.0 # channel half length
    hgt = 3.0 # channel half height
    wdt = 3.0 # channel width
    S_ref = 1.*spn # reference area
    c_ref = 1 # reference chord
    fms = 1.0 # farfield mesh size
    nms = 0.05 # nearfield mesh size

    # parameters for mesh deformation
    alfa = 3*np.pi/180
    xc = 0.25
    dz_max = 0.2*spn

    # mesh an airfoil
    print(ccolors.ANSI_BLUE + 'PyMeshing...' + ccolors.ANSI_RESET)
    tms['msh'].start()
    pars = {'spn' : spn, 'lgt' : lgt, 'wdt' : wdt, 'hgt' : hgt, 'msLe' : nms, 'msTe' : nms, 'msF' : fms}
    msh, gmshWriter = floD.mesh(dim, 'models/n0012_3.geo', pars, ['field', 'wing', 'symmetry', 'downstream'], wktp = 'wakeTip')
    tms['msh'].stop()
    # create mesh deformation handler
    mshDef = floD.meshMorpher(msh, dim, ['wing'])

    # set the problem and add medium, initial/boundary conditions
    tms['pre'].start()
    pbl, _, _, bnd, _ = floD.problem(msh, dim, alpha0, 0., M_inf, S_ref, c_ref, 0., 0., 0., 'wing', tp = 'teTip')
    tms['pre'].stop()

    # solver problem for 0°AOA
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver0'].start()
    solver = floD.newton(pbl)
    solver.run()
    tms['solver0'].stop()

    # deform the mesh (dummy rotation + bending)
    print(ccolors.ANSI_BLUE + 'PyDeforming...' + ccolors.ANSI_RESET)
    tms['deform'].start()
    mshDef.savePos()
    rMat = np.matrix([[np.cos(alfa), np.sin(alfa)], [-np.sin(alfa), np.cos(alfa)]])
    vNods = {}
    for e in bnd.groups[0].tag.elems:
        for n in e.nodes:
            if n.no not in vNods:
                vNods[n.no] = n
    for no in vNods:
        nod = np.matrix([[vNods[no].pos[0]-xc], [vNods[no].pos[2]]])
        new = rMat * nod
        vNods[no].pos[0] = new[0,0]
        vNods[no].pos[2] = new[1,0] + dz_max/(spn*spn)*(vNods[no].pos[1]*vNods[no].pos[1])
    # deform the mesh
    mshDef.deform()
    gmshWriter.save(msh.name+"_def")
    tms['deform'].stop()

    # solve problem for alfa AOA
    print(ccolors.ANSI_BLUE + 'PySolving...' + ccolors.ANSI_RESET)
    tms['solver1'].start()
    solver.run()
    solver.save(gmshWriter)
    tms['solver1'].stop()

    # display results
    print(ccolors.ANSI_BLUE + 'PyRes...' + ccolors.ANSI_RESET)
    print('  case        M    alpha       Cl       Cd       Cm')
    print('  true {0:8.2f} {1:8.1f} {2:8.4f} {3:8.4f} {4:8.4f}'.format(M_inf, alfa*180/np.pi, solver.Cl, solver.Cd, solver.Cm))

    # display timers
    tms['total'].stop()
    print(ccolors.ANSI_BLUE + 'PyTiming...' + ccolors.ANSI_RESET)
    print('CPU statistics')
    print(tms)

    # check results
    print(ccolors.ANSI_BLUE + 'PyTesting...' + ccolors.ANSI_RESET)
    tests = CTests()
    if alfa == 3*np.pi/180 and M_inf == 0.3 and spn == 1.0:
        tests.add(CTest('iteration count', solver.nIt, 3, 1, forceabs=True))
        tests.add(CTest('CL', solver.Cl, 0.135, 5e-1))
        tests.add(CTest('CD', solver.Cd, 0.0062, 1e-2)) # Tranair (NF=0.0062, FF=0.0030), Panair 0.0035
        tests.add(CTest('CS', solver.Cs, -0.011, 5e-2))
        tests.add(CTest('CM', solver.Cm, 0.0061, 1e-2))
    tests.run()

    # eof
    print('')

if __name__ == "__main__":
    main()
